# -*- coding: utf-8 -*-
"""
Created on Wed Nov  7 13:33:06 2018

@author: compactmatter
"""


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
plt.switch_backend('agg')
from sigproSS import signatures_optimizer as sigopt
import os
import shutil
import multiprocessing as mp
import xml.etree.ElementTree as ET  
from lxml import etree
from SigProfilerMatrixGenerator.scripts import SigProfilerMatrixGeneratorFunc as matGen
from sigproextractor import subroutines as sub
import sigproSS as cosmic
import sigProfilerPlotting as plot


def get_connected_alwaysinclude_signatures(signatureRulesXML, signatures):
    tree = ET.parse(signatureRulesXML)      
    root = tree.getroot()
    
    #Connected
    csList = [] 
    csIDs = []      
    setList = []
    csLength = []
    connectedSignatures = []
    connectedSignaturesIDs = []

    for conn in root.find('connectedSignatures').findall('connected'):
        csLength.append(len(conn))
        setList.append(conn.get("Set"))
        for i in conn.findall('signature'):
            csList.append(i.text)

    csLengthCumSum = np.cumsum(csLength)
    for j in range(len(csList)):
        csIDs.append(signatures.index(csList[j]))
    for i in range(len(csLength)):
        if i == 0:
            connectedSignatures.append(csList[0:csLengthCumSum[i]])
            connectedSignaturesIDs.append(csIDs[0:csLengthCumSum[i]])
        else:
            connectedSignatures.append(csList[csLengthCumSum[i-1]:csLengthCumSum[i]])
            connectedSignaturesIDs.append(csIDs[csLengthCumSum[i-1]:csLengthCumSum[i]])
    
    #Always Include
    allowEverywhere = []
    for ai in root.find('alwaysIncludeSignatures').findall('signature'):
        allowEverywhere.append(ai.text)
    
    allowEverywhereIDs = [i for i, x in enumerate(signatures) if x in allowEverywhere]
    
    return[connectedSignaturesIDs, allowEverywhereIDs]
    


def check_signature_rules(signaturesInSample,
                          signatures,
                          sampleID,
                          seqType,
                          totalMutations,
                          C_to_A_p,
                          C_to_A_d,
                          C_to_G_p,
                          C_to_G_d,
                          C_to_T_p,
                          C_to_T_d,
                          T_to_A_p,
                          T_to_A_d,
                          T_to_C_p,
                          T_to_C_d,
                          T_to_C_ATN_p,
                          T_to_C_ATN_d,
                          T_to_G_p,
                          T_to_G_d,
                          signatureRulesXML,
                          signatureRulesXMLSchema):
    
    totalSignatures = len(signaturesInSample)  
    
    
    tree = ET.parse(signatureRulesXML)  
    
    #Validate XML file against schema
    xml = etree.parse(signatureRulesXML)

    schema_doc = etree.parse(signatureRulesXMLSchema)
    schema = etree.XMLSchema(schema_doc)

    if (schema.validate(xml) == False):
        print(schema.assertValid(xml))
    
    root = tree.getroot()

    signaturesList = []    
    for i in range(totalSignatures):
        for signature in root.findall('signatureName'):
            signaturesList.append(signature.get("signatureSBS"))
            
        if(signatures[i] in signaturesList):
            if (root.find(".//strandBias/..[@signatureSBS='" + signatures[i] + "']") is not None):
                strandbias = root.find(".//strandBias/..[@signatureSBS='" + signatures[i] + "']").find('strandBias').findall('mutationType')
                for sb in range(len(strandbias)):
                    if( eval(strandbias[sb].get('mutType').replace('>','_to_') + "_p") > float(strandbias[sb][1].text) or
                       eval(strandbias[sb].get('mutType').replace('>','_to_') + "_d") != float(strandbias[sb][0].text)):
                       signaturesInSample[i] = 0
                       
            if (root.find(".//totalMutations/..[@signatureSBS='" + signatures[i] + "']") is not None):
                totalmutations = root.find(".//totalMutations/..[@signatureSBS='" + signatures[i] + "']").find('totalMutations').findall('seqType')
                for st in range(len(totalmutations)):
                    if( totalmutations[st].get('type') == seqType):
                        if( float(totalMutations) < float(totalmutations[st][0].text) ):
                            signaturesInSample[i] = 0
                
    return signaturesInSample



def parallel_for_loop(iSample, inputSamples, allGenomeSignatures, allExomeSignatures, cancerType,
                      seqType, totalMutations, sampleNames, samplePvalue,
                      longSampleNames, totalSignatures, useRules, sigNames, idsToAdd, connected,
                      exposuresNew, accr_org, kl_div_org, frob_rel_div_org, norm_one_dif_org):
    
    
    
    #get the path for files
    paths = cosmic.__path__[0]
    
    genome = inputSamples[:,iSample]
    genome = genome.reshape(96,1) #Uncomment for other functions
    if (seqType[iSample] == 'WGS'):
        allSignatures =  allGenomeSignatures
    else:
        allSignatures =  allExomeSignatures
        
    #Identify signatures in cancer type and apply signature rules
    
    
                  

    longSampleName = cancerType[iSample] + '::' + sampleNames[iSample]

    signaturesInSamples = np.zeros([allSignatures.shape[1], inputSamples.shape[1]])
        
    exposures = signaturesInSamples[:, longSampleNames.index(longSampleName)]
    #print(exposures.shape)
    if ( exposures.size == 0 ):
        exposures = np.zeros((totalSignatures, 1))
    
      
    
    C_to_A_p = samplePvalue.iloc[0][0]
    C_to_A_d = samplePvalue.iloc[0][1]
    C_to_G_p = samplePvalue.iloc[1][0]
    C_to_G_d = samplePvalue.iloc[1][1]
    C_to_T_p = samplePvalue.iloc[2][0]
    C_to_T_d = samplePvalue.iloc[2][1]
    T_to_A_p = samplePvalue.iloc[3][0]
    T_to_A_d = samplePvalue.iloc[3][1]
    T_to_C_p = samplePvalue.iloc[4][0]
    T_to_C_d = samplePvalue.iloc[4][1]
    T_to_C_ATN_p = samplePvalue.iloc[5][0]
    T_to_C_ATN_d = samplePvalue.iloc[5][1]
    T_to_G_p = samplePvalue.iloc[6][0]
    T_to_G_d = samplePvalue.iloc[6][1]
    
    if ( useRules == True ):
        #print("Using rules")
        exposures = check_signature_rules(exposures,
                                          sigNames,
                                          iSample,
                                          seqType,
                                          totalMutations,
                                          C_to_A_p,
                                          C_to_A_d,
                                          C_to_G_p,
                                          C_to_G_d,
                                          C_to_T_p,
                                          C_to_T_d,
                                          T_to_A_p,
                                          T_to_A_d,
                                          T_to_C_p,
                                          T_to_C_d,
                                          T_to_C_ATN_p,
                                          T_to_C_ATN_d,
                                          T_to_G_p,
                                          T_to_G_d,
                                          paths+'/input/20181108_Signature_Rules.xml',
                                          paths+'/input/Signature_Rules_Schema.xsd')
    
    
    #print("Total Mutations", totalMutations[0])
    
    #Remove all signatures in the sample: one by one
    #Add signatures that are allowed everywhere    
    if (idsToAdd != False):
        exposures[idsToAdd] = 1 
    
    #Add connected signatures
    if (len(connected) > 0):
        for iConnect in range(len(connected)):
            if ( sum(exposures[connected[iConnect]]) > 0 ):
                exposures[connected[iConnect]] = 1
    
    #print(exposures)
    ##########################################################[exposures, accr, kl_div, frob_rel_div, norm_one_dif] = remove_all_single_signatures(exposures, allSignatures, genome, sigNames)
    
    
    IDs = [i for i,x in enumerate(exposures) if x > 0]
    #!print("Signatures Selected to Remove after checking the rules:")
    #!print(IDs)
    processes = allSignatures[:,IDs]

    exposure =   sigopt.initial_optimization(processes, genome) #processes, sample 
    #exposure, similarity =   sigopt.fit_signatures(processes, genome) #processes, sample
    #exposure = exposure[:,np.newaxis]

    pool = mp.Pool(processes=1)
    results = [pool.apply_async(sigopt.remove_signatures, args=(x,processes,exposure,genome,)) for x in range(genome.shape[1])]
    
    output = [p.get() for p in results]
    pool.close()
    
    exposureAfterRemovingSignatures = np.zeros(exposure.shape)
    for i in range(len(output)):
        exposureAfterRemovingSignatures[:,i]=output[i] 

    np.put(exposures, IDs, exposureAfterRemovingSignatures.ravel())
    #print(np.nonzero(exposures)[0])
    
    #Add all remaining signatures to the sample: one by one
    #Add signatures that are allowed everywhere
    #print(exposures)
    
    if (idsToAdd != False):
        for i in idsToAdd:
            exposures[i] = 1
       
    #Add connected signatures
    if (len(connected) > 0):
        for iConnect in range(len(connected)):
            if ( sum(exposures[connected[iConnect]]) > 0 ):
                exposures[connected[iConnect]] = 1

    #print(exposures)
    IDs = [i for i,x in enumerate(exposures) if x > 0] #Changed i to i+1 
    #print(IDs)
    
    #print(exposureAfterRemovingSignatures) ###############
    
    
    ######################################################exposures = add_all_single_signatures(exposures, allSignatures, genome, sigNames) #No eval single sample now
    ## ## ## ##processes = allSignatures[:,IDs]
    
    ####################print(IDs) 
    #pd.DataFrame(processes).to_csv("processes_file.txt", sep = "\t")
    #pd.DataFrame(genome).to_csv("genome_file.txt", sep = "\t")
    #print("done")
    
    exposureAfterAddingSignatures = sigopt.add_signatures(allSignatures, genome, 0.025, IDs)[0]
    #print(len(exposureAfterAddingSignatures))
    #Add signatures that are allowed everywhere
    exposures = exposureAfterAddingSignatures
    ## ## ## ## np.put(exposures, IDs, exposureAfterAddingSignatures.ravel())
   
    
    
    #print(exposures)#####################
    
    
    if (idsToAdd != False):
        exposures[idsToAdd] = 1

    #Add connected signatures
    if (len(connected) > 0):
        for iConnect in range(len(connected)):
            if ( sum(exposures[connected[iConnect]]) > 0 ):
                exposures[connected[iConnect]] = 1
    
    #print(exposures)
    IDs = [i for i,x in enumerate(exposures) if x > 0]
   
    
    ################################################################[exposures, accr_org[iSample], kl_div_org[iSample], frob_rel_div_org[iSample], norm_one_dif_org[iSample]] = eval_single_sample(exposures, allSignatures, genome)

    #IDs = [i for i,x in enumerate(exposures) if x > 0] #BEWARE
    #processes = allSignatures[:,IDs]#BEWARE

    #exposure =   initial_optimization(processes, genome)#BEWARE
    #exposures = np.zeros((totalSignatures, 1)) #BEWARE
    #np.put(exposures, IDs, exposure.ravel()) #BEWARE
    #print(exposures)
    exposuresNew[:, iSample] = exposures.ravel() #exposures
    #print(exposuresNew)
    indices = np.nonzero(exposuresNew)[0] 
    #print(exposuresNew[list(indices)])
    
    
    #fit the signatures 
    print("\n")
    sigNames=np.array(sigNames)
    exposures, similarity = sigopt.fit_signatures(allSignatures[:,list(indices)], genome)
    
    
    
    
    
    #Dispay summary output data
    #out_str = ['Sample #' + str(iSample+1) + ': ' + str(inputSamples['cancerType'][iSample][0]) + ' ' + str(inputSamples['sampleNames'][iSample][0]) + ' with ' + str(sum(exposuresNew[:, iSample]>0)) + ' signatures and an accuracy of ' + str(round(accr_org[iSample][0],2))] #BEWARE ,'%.2f' removed
    #out_str = ['Sample #' + str(iSample+1) + ': [' + str(cancerType[iSample]) + '] [' + str(sampleNames[iSample]) + '] with ' + str(sum(exposuresNew[:, iSample]>0)) + ' signatures']
    
    print(indices)
    print(exposures)
    print(sigNames)
    print(allSignatures.shape)
    print(similarity)
    return([indices,exposures,sigNames, allSignatures, similarity])               


def analysis_individual_samples(#signaturesSet,
                                #signaturesInSamples,
                                samples,
                                #outputFolder,
                                #outputFile,
                                useRules,
                                #allowSigsEverywhere,
                                #connected,
                                newSignatures_file,
                                genomeSignatures_file,
                                exomeSignatures_file,
                                sampleNames,
                                sampleCancerTypes,
                                cancerType,
                                seqType,
                                totalMutations,
                                p_value,
                                signatureRulesXML):
    
   
    #Analysis of data
    #Loading samples
    
    inputSamples = samples
    
    
    
    #print(inputSamples)
    #print("\n")
    try:
        totalSamples = inputSamples.shape[1] 
    except:
        inputSamples = inputSamples[:,np.newaxis]
        totalSamples = inputSamples.shape[1] 
        
    accr_org = np.zeros((totalSamples, 1))
    kl_div_org = np.zeros((totalSamples, 1))  
    frob_rel_div_org = np.zeros((totalSamples, 1))
    norm_one_dif_org = np.zeros((totalSamples, 1)) 
    #print(inputSamples)
    #Loading signatures
    newSignatures = open(newSignatures_file,'r').read().split('\n')
    

    
    allGenomeSignatures = np.loadtxt(genomeSignatures_file)  
    allExomeSignatures  = np.loadtxt(exomeSignatures_file)
    sigNames = newSignatures
    totalSignatures = allGenomeSignatures.shape[1]
    exposuresNew = np.zeros((totalSignatures, totalSamples))
    #Loading signatures in samples
    
    
    
    
    
    longSampleNames = ["" for x in range(len(sampleNames))]
    
    for i in range(len(sampleNames)): #MATLAB CODE for i = 1 : size(sigsInCanType.sampleNames, 1)
       longSampleNames[i] = sampleCancerTypes[i] + '::' + sampleNames[i]

    

    #print(Parallel(n_jobs = num_cores)(delayed(parallel_for_loop)(iSample, inputSamples, allGenomeSignatures, allExomeSignatures, cancerType_file, seqType_file, totalMutations_file, sampleNames, signaturesInSamples_file, longSampleNames, totalSignatures, useRules, sigNames, idsToAdd, connected, saOptions, exposuresNew, accr_org, kl_div_org, frob_rel_div_org, norm_one_dif_org) for iSample in range(totalSamples)))
    #print(sigNames)
    [connected, allowSigsEverywhere] = get_connected_alwaysinclude_signatures(signatureRulesXML, sigNames)
    #print(connected)
    #print(allowSigsEverywhere)
    
    if(dir().count('allowSigsEverywhere') > 0 and len(allowSigsEverywhere) > 0):
        ###!print('Allow these signatures in all samples:')
        for i in range(len(allowSigsEverywhere)):
            ###!print(newSignatures[allowSigsEverywhere[i]])
            idsToAdd = allowSigsEverywhere
    else:
        idsToAdd = False 
    
    if (dir().count('connected') > 0 and len(connected) > 0):
        ###!print('These mutational signatures are connected:')
        for i in range(len(connected)):
            dispStr = 'Set ' + str(i+1) + ':' + newSignatures[connected[i][0]]
            for j in range(1, len(connected[i])):
                dispStr = dispStr + '; ' + newSignatures[connected[i][j]]
            ###!print(dispStr)

    
    for iSample in range(totalSamples):
        samplePvalue = p_value.iloc[:,iSample]
        results = parallel_for_loop(iSample, inputSamples, allGenomeSignatures, allExomeSignatures, cancerType,
                                seqType, totalMutations, sampleNames, samplePvalue,
                                longSampleNames, totalSignatures, useRules, sigNames, idsToAdd, connected
                                , exposuresNew, accr_org, kl_div_org, frob_rel_div_org, norm_one_dif_org)
        
    
        return results


def decomposition_profile(exposures, similarity, signatureList , sampleName):

  
    
    signames = signatureList 
    
    
    exposure_percentages = exposures[np.nonzero(exposures)]/np.sum(exposures[np.nonzero(exposures)])*100
    listofinformation = list("0"*len(np.nonzero(exposures)[0])*3)
    
    count =0
    for j in np.nonzero(exposures)[0]:
        listofinformation[count*3] = signames[j]
        listofinformation[count*3+1] = round(exposure_percentages[count],2)
        listofinformation[count*3+2]="%"
        count+=1
    ListToTumple = tuple([sampleName] +listofinformation+ [similarity])
    strings ="Sample %s,"+" %s (%0.2f%s) &"*(len(np.nonzero(exposures)[0])-1)+" %s (%0.2f%s), %0.2f\n" 
    return(strings%(ListToTumple))
       



def single_sample(vcf, outputdir, ref="GRCh37", exome=False):
    
    
    """
    Decompose the query samples into the global signatures.
    
    parameters
    ----------
    vcf: A string. The name of the folder containing the vcf files. The folder should be present in the current working directory. 
    outputdir: A string. The name of the output folder. The output folder will be generated in the current working directory according to name provided in the current working directory. 
    ref: A string. The name of the reference genome file. The file should be installed previously through "SigProfilerMatrixGenerator". 
    Please see the "INSTALLATION" part of the README.md file. The default reference genome is "GRCh37".
    exome: A boolean. If the agrument is True, that will genearate the mutational profile only for the exomes. If False, the profile 
    for the whole genome sequence will be generated. 
          
    
    Returns:
    -------
    After the single_sample function is successfully executed, an output directory will be generated in the current working directory. 
    
    The output folder will contain the following files:
        -exposure.txt 
        -signature.txt 
        -probabilities.txt 
        -signature plot pdf 
        -dendrogram plot
        -decomposition profile.csv

    Example: 
    -------
    >>> from sigproSS import spss 
    >>> data = spss.importdata()
    >>> spss.single_sample(data, "results", ref="GRCh37", exome=False)
    
    
        
    """
    
    if not os.path.exists(outputdir):
        os.makedirs(outputdir)
        
        
    
    #get the path for files
    paths = cosmic.__path__[0]
    
    if vcf[-1] != "/":
        vcf_name = vcf.split("/")[-1]
    else:
        vcf_name = vcf.split("/")[-2]
    

    data = matGen.SigProfilerMatrixGeneratorFunc(vcf_name, ref, vcf, exome=exome)
    
    # make the totalExposure dataframe which have dimention of totalsignatures and totalsamples
    
    totalExposures = np.zeros([65,data["96"].shape[1]])
    listOfSamples = list(data["96"].head(0))
    
    # open a file to profile the signatures
    fh = open(outputdir+"/decomposition profile.csv", "w")
    fh.write("Sample Names, Global NMF Signatures, Similarity\n")
    fh.close()
    
    for i in range(data["96"].shape[1]):
        print("##########################################################")
        print("Exacting Profile for "+"Sample " +str(i+1))
        index = i
        samples = data["96"].iloc[:,index:index+1]
        p_value = data["7_pvalue"]
        #print(p_value)
        samples = np.array(samples)
        sampleNames = list(data["96"].head(0))[index:index+1]
        cancerType = ['Breast Cancer']*samples.shape[1]
        seqType = ['WGS']*samples.shape[1]
        totalMutations= np.sum(samples, axis=0)
        
      
        
        
        #results variable contains [indices,exposures, signatureNames, allSignatures, similarity]
        results = analysis_individual_samples(samples, 
                                    True,
                                    paths+'/input/signaturesSet.txt',
                                    paths+'/input/genomeSignatures.txt',
                                    paths+'/input/exomeSignatures.txt',
                                    sampleNames,
                                    cancerType,
                                    cancerType,
                                    seqType,
                                    totalMutations,
                                    p_value,
                                    paths+'/input/20181108_Signature_Rules.xml') 
        totalExposures[results[0],i]=results[1]
        listOfSignatures = results[2]
        signatures = pd.DataFrame(results[3])
        profile = decomposition_profile(totalExposures[:,i],  results[4], results[2], sampleNames[0])
        
        #write the profiles into file
        fh = open(outputdir+"/decomposition profile.csv", "a")
        fh.write(profile)
        fh.close()
        
        
    
    
    
    
    #prepare the exposures dataframe
    
    totalExposures = pd.DataFrame(totalExposures)
    totalExposures = totalExposures.set_index(listOfSignatures)
    totalExposures.columns = listOfSamples
    totalExposures = totalExposures.rename_axis("Samples", axis="columns")
    #Convert the floats to integers
    totalExposures[listOfSamples] = totalExposures[listOfSamples].applymap(np.int64)
    
    
    
    #remove the rows with all zeros to create the final exposure dataframe
    exposures = totalExposures.loc[~(totalExposures==0).all(axis=1)]
    
    #presure the signatures dataframe
    signatures = pd.DataFrame(results[3])
    signatures.columns = listOfSignatures
    signatures = signatures.set_index(data["96"].index)
    signatures = signatures.rename_axis("Signatures", axis="columns")
    
    #Filter the signatures by the exposures rows to get the final signature dataframe
    signatures = signatures.loc[:,list(exposures.index)] 
    
    #create the probalities 
    probability = sub.probabilities(signatures, exposures)
    probability = probability.set_index("Sample")
    probability = probability.rename_axis("", axis="columns")
    
    try:
        #create the dedrogrames
        Y, dn = sub.dendrogram(exposures, 0.05, outputdir)
    except:
        pass
        
    #export results
    signatures.to_csv(outputdir+"/signatures.txt", "\t", index_label=[signatures.columns.name]) 
    exposures.to_csv(outputdir+"/exposures.txt", "\t", index_label=[exposures.columns.name]) 
    probability.to_csv(outputdir+"/probabilities.txt", "\t")
    plot.plotSBS(outputdir+"/signatures.txt", outputdir+"/Signature_plot", "", "96", True)

def importdata():
    
    """
    Imports the path of example data.
    
    parameters
    ----------
            
    
    Returns:
    -------

    The path of the example data.

    Example: 
    -------
    >>> from sigproSS import spss 
    >>> data = spss.importdata()
    
    This "data" variable can be used as a parameter of the first argument of the sigpross function.
        
    """
    
    
    
    paths = cosmic.__path__[0]
    directory = os.getcwd()
    dataold = paths+"/input/vcf"
    datanew = directory+"/vcf"
    if not os.path.exists(datanew):
        shutil.copytree(dataold , datanew) 
    data="vcf"
    return data





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb 14 15:22:11 2019

@author: mishugeb
"""


import os
import warnings
import pandas as pd
import numpy as np
import copy
import time
from scipy.stats import binom_test
import multiprocessing
from functools import partial
from sigproSS import signatures_optimizer as sigopt
from sigproSS import spss
from sigproextractor import subroutines as sub
import sigproSS as cosmic

mutation_categories = {
    'SBS':['C>A','C>G','C>T','T>A','T>C','T>G'],
    'DBS':['AC>','AT>','CC>','CG>','CT>','GC>','TA>','TC>','TG>','TT>'],
    'ID':['DEL_C_1', 'DEL_T_1', 'INS_C_1', 'INS_T_1', 'DEL_repeats_2', 'DEL_repeats_3', 'DEL_repeats_4', 'DEL_repeats_5+',
    'INS_repeats_2', 'INS_repeats_3', 'INS_repeats_4', 'INS_repeats_5+', 'DEL_MH_2', 'DEL_MH_3', 'DEL_MH_4', 'DEL_MH_5+'],
}

# each signature corresponds to a dictionary with categories and relevant TSB signs:
# -1 means a negative TSB, i.e. more mutations on untranscribed strand,
# 1 means a positive TBS, i.e. more mutations on transcribed strand

def get_indeces(a, b):
    
    """ 
    Extracts the indices multiple items in a list.
    
    Parameters:
        a: list. where we want to get the index of the items.
        b; list. the items we want to get index of. 
    """

    indeces = []
    for i in b:
        try: 
            idx = a.index(i)
            indeces.append(idx)
        except: 
            next

    return indeces 

    """
    #example: 
    x = [1,3,5,8]
    y = [3, 8]
    get_indeces(x, y)
    #result
    >>> [1,3]
    """
    

def get_items_from_index(x,y):
    """ decipher the values of items in a list from their indices.
    """
    z = []
    for i in y:
        z.append(x[i])
    return z

    """
    #example: 
    x = [1,3,5,8]
    y = [1, 3]
    get_items_from_index(x, y)
    #result
    >>> [3,8]
    """
    
#############################################################################################################
##################################### Get Input From CSV Files ##############################################
#############################################################################################################
def read_csv(filename, folder = False):
    
    # if filename is not a string then get the genomes from the object
    if type(filename) != str:
        genomes = filename
        
    else:
        
        if folder==False:
            genomes = pd.read_csv(filename, sep=",").iloc[:, :]    
        else:
        
            count = 1
            for file in os.listdir(filename):
                #print(count) 
                df = pd.read_csv(filename+"/"+file)
                if count==1:
                        genomes = df
                else:
                        genomes = pd.merge(genomes, df, on=["Mutation type", "Trinucleotide"])
                count += 1


    

    
    
    mtypes = [str(genomes.shape[0])]
    
    if mtypes == ['96']:
        # create the list to sort the mutation types
        orderlist = ['A[C>A]A', 'A[C>A]C', 'A[C>A]G', 'A[C>A]T', 'A[C>G]A', 'A[C>G]C', 'A[C>G]G', 'A[C>G]T', 'A[C>T]A', 'A[C>T]C', 'A[C>T]G', 'A[C>T]T',
                     'A[T>A]A', 'A[T>A]C', 'A[T>A]G', 'A[T>A]T', 'A[T>C]A', 'A[T>C]C', 'A[T>C]G', 'A[T>C]T', 'A[T>G]A', 'A[T>G]C', 'A[T>G]G', 'A[T>G]T',
                     'C[C>A]A', 'C[C>A]C', 'C[C>A]G', 'C[C>A]T', 'C[C>G]A', 'C[C>G]C', 'C[C>G]G', 'C[C>G]T', 'C[C>T]A', 'C[C>T]C', 'C[C>T]G', 'C[C>T]T',
                     'C[T>A]A', 'C[T>A]C', 'C[T>A]G', 'C[T>A]T', 'C[T>C]A', 'C[T>C]C', 'C[T>C]G', 'C[T>C]T', 'C[T>G]A', 'C[T>G]C', 'C[T>G]G', 'C[T>G]T',
                     'G[C>A]A', 'G[C>A]C', 'G[C>A]G', 'G[C>A]T', 'G[C>G]A', 'G[C>G]C', 'G[C>G]G', 'G[C>G]T', 'G[C>T]A', 'G[C>T]C', 'G[C>T]G', 'G[C>T]T',
                     'G[T>A]A', 'G[T>A]C', 'G[T>A]G', 'G[T>A]T', 'G[T>C]A', 'G[T>C]C', 'G[T>C]G', 'G[T>C]T', 'G[T>G]A', 'G[T>G]C', 'G[T>G]G', 'G[T>G]T',
                     'T[C>A]A', 'T[C>A]C', 'T[C>A]G', 'T[C>A]T', 'T[C>G]A', 'T[C>G]C', 'T[C>G]G', 'T[C>G]T', 'T[C>T]A', 'T[C>T]C', 'T[C>T]G', 'T[C>T]T',
                     'T[T>A]A', 'T[T>A]C', 'T[T>A]G', 'T[T>A]T', 'T[T>C]A', 'T[T>C]C', 'T[T>C]G', 'T[T>C]T', 'T[T>G]A', 'T[T>G]C', 'T[T>G]G', 'T[T>G]T']
        #Contruct the indeces of the matrix
        #setting index and columns names of processAvg and exposureAvg
        index1 = genomes.iloc[:,1]
        index2 = genomes.iloc[:,0]
        index = []
        for i, j in zip(index1, index2):
            index.append(i[0]+"["+j+"]"+i[2])
    
    
        index = np.array(pd.Series(index))
        genomes["index"] = index
    
    
        #sort the mutationa types
        genomes["index"]= pd.Categorical(genomes["index"], orderlist)
        genomes = genomes.sort_values("index")
        
    
        genomes = genomes.iloc[:,2:genomes.shape[1]]
    
    
    
        # set the index 
        genomes = genomes.set_index("index")
    
        # prepare the index and colnames variables
        index = np.array(orderlist)
        colnames = np.array(pd.Series(genomes.columns.tolist()))
    
    else:
        
        index = np.array(genomes.iloc[:,0])
        genomes = genomes.iloc[:,1:]
        genomes = genomes.loc[:, (genomes != 0).any(axis=0)]
        colnames = genomes.columns
       
        
        
   
    
    
    return(genomes, index, colnames, mtypes)


def make_folder_if_not_exists(folder):
    if not os.path.exists(folder):
        try:
            os.makedirs(folder)
        except:
            warnings.warn("Could not create a folder ", folder)

def calculate_strand_bias(list_of_values):
    """
    Perform a test to compare two Poisson means (mutation counts on transcribed
    and untranscribed strands) usng binom_test function from scipy.stats.
    Here, an exact binomial test is used which is preferred for low mutation counts.

    Parameters:
    ----------
    list_of_values: array_like
        list of two values, e.g. [23, 32]

    Returns:
    -------
    The p-value of the hypothesis test
    -------
    """
    if len(list_of_values)!=2:
        raise ValueError("Please provide a list of two values (transcribed and untranscribed strands counts)")

    return binom_test(list_of_values)

def annotate_strand_bias(input_data, p_value=0.05):
    #print("###################  INPUT DATA ########################")
    #print(input_data)
    """
    Perform a test of strand bias on each row of an input dataframe.

    Parameters:
    ----------
    input_data: pandas dataframe
        input dataframe that has exactly two columns (transcribed and untranscribed strand mutation counts)
    p_value: float, optional
        threshold p-value to deem the strand bias significant. Default is 0.05

    Returns:
    -------
    dataframe, strand_bias_flag, affected_categories: tuple

    dataframe: pandas dataframe
    Dataframe with star-marked rows with significant TSB

    strand_bias_flag: boolean
    Boolean flag showing if significant TSB is observed in any row

    affected_categories: dict
        Dictionary of row indexes with significant TSB with absolute differences (transcribed minus untransribed)
    -------
    """
    dataframe = copy.deepcopy(input_data)
    strand_bias_flag = False
    if not isinstance(dataframe, pd.DataFrame):
        raise ValueError("Input not a dataframe, can not annotate for strand bias. Given type:",type(dataframe))
    affected_categories = {}
    for category in list(dataframe.index):
        transcribed_strand_count, untranscribed_strand_count = dataframe.loc[category].tolist()
        strand_bias_p_value = calculate_strand_bias([transcribed_strand_count, untranscribed_strand_count])
        if strand_bias_p_value < p_value:
            # check which strand has more mutations
            difference = transcribed_strand_count-untranscribed_strand_count
            dataframe.rename(index={category: '$\star$'+category}, inplace=True)
            strand_bias_flag = True
            affected_categories[category] = difference
    #print("################### DATA FRAME ########################")
    #print(dataframe)
    #print("################### strand_bias_flag ########################")
    #print(strand_bias_flag)
    #print("################### affected_categories ########################")
    #print(affected_categories)
    return dataframe, strand_bias_flag, affected_categories

def rearrange_stranded_data(input_data):
    
    """
    Rearrange stranded data of a single sample into two different columns of the dataframe

    Parameters:
    ----------
    input_data: pandas dataframe
        Input single-column dataframe with a 3-level MultiIndex of 192-context PCAWG format (e.g. WGS_PCAWG.192.csv on Synapse)
        Essential to have a 'Strand' level of the MultiIndex, with 'T' and 'U' values

    Returns:
    -------
    rearranged_data: pandas dataframe
        Dataframe with removed 'Strand' level, but with two columns (transcribed and untranscribed strands)
    -------
    """
    data = copy.deepcopy(input_data)
    
    T_strand, U_strand = pd.DataFrame(), pd.DataFrame()
    T_strand = data.iloc[data.index.get_level_values('Strand') == 'T']
    U_strand = data.iloc[data.index.get_level_values('Strand') == 'U']
    T_strand = T_strand.droplevel(0)
    U_strand = U_strand.droplevel(0)
    rearranged_data = pd.concat([T_strand, U_strand], names=['transcribed strand','untranscribed strand'], axis=1, sort=True)
    rearranged_data.columns = ['transcribed strand','untranscribed strand']
    return rearranged_data

def categorise_mutations(input_data, categories, condensed=False, strand_bias=False):
    """
    Categorise the dataframe by a given a list of categories

    Parameters:
    ----------
    input_data: pandas dataframe
        Input dataframe. If strand_bias flag is True, a single-column dataframe is expected with
        a 3-level MultiIndex of 192-context PCAWG format (see rearrange_stranded_data function)

    categories: list of strings
        Mutation categories to condense mutations to, e.g.: ['C>A','C>G','C>T','T>A','T>C','T>G']

    condensed: boolean
        If True, the function returns a dataframe with mutations summed up for given categories,
        otherwise, a dictionary of dataframes is returned with keys for each category

    strand_bias: boolean
        If True, rearrange_stranded_data function is called on a dataframe to obtain a two-column format.

    Returns:
    -------
    categorised_mutations: pandas dataframe
        Dataframe with categorised mutations if condensed flag is True, otherwise
        a dictionary of dataframes with keys for each category listed in input categories
    -------
    """
  
    
    data = copy.deepcopy(input_data)
    if strand_bias:
        data = rearrange_stranded_data(data)

    # dataframe in case of condensed format, dictionary of dataframes otherwise
    if condensed:
        if strand_bias:
            categorised_mutations = pd.DataFrame(columns=['transcribed strand','untranscribed strand'])
        else:
            if isinstance(data, pd.Series):
                categorised_mutations = pd.DataFrame(columns=['N'])
            else:
                categorised_mutations = pd.DataFrame(columns=data.columns)
    else:
        categorised_mutations = {}

    for category in categories:
        mutations_per_category = data.filter(like=category, axis=0)

        # sum up categories for condensed format, otherwise leave it in dictionary
        if condensed:
            categorised_mutations.loc[category] = mutations_per_category.sum()
        else:
            categorised_mutations[category] = mutations_per_category
    
    
    return categorised_mutations

def apply_TSB_rules(input_mutations, input_signatures, mutation_type, verbose=False):
    
    #print("###################  input_mutations ########################")
    #print(input_mutations)
    #print("###################  input_signatures ########################")
    #print(input_signatures)
    #print("###################  mutation_type ########################")
    #print(mutation_type)
    """
    Apply transcriptional strand bias rules for a single-column dataframe (single sample)
    If a strand bias is observed in a sample but not a signature, it is excluded from the input list
    A 'sign' of TSB is respected (has to be the same in both samples and signatures)

    Parameters:
    ----------
    input_mutations: pandas dataframe
        Input dataframe. A single-column dataframe is expected with a 3-level MultiIndex
        of 192-context PCAWG format (see rearrange_stranded_data function)

    input_signatures: list of strings
        List of input signatures to apply the rules for

    mutation_type: string
        Mutation type (e.g. SBS, DBS, ID), used to pull the list of relevant categories

    verbose: boolean
        Verbosity flag: lots of output for debugging if set to True

    Returns:
    -------
    signatures_after_TSB_rules: list of strings
        The list of signatures passing the strand bias rules
    -------
    """
    signatures_with_TSB = {
    'SBS4':{'C>A':1,'T>A':1},
    'SBS5':{'T>C':1},
    'SBS7a':{'C>T':-1},
    'SBS7b':{'C>T':-1},
    'SBS7c':{'T>A':-1},
    'SBS7d':{'T>C':-1},
    'SBS8':{'C>A':1},
    'SBS16':{'T>C':1},
    'SBS19':{'C>T':1},
    'SBS21':{'T>C':-1},
    'SBS22':{'T>A':1},
    'SBS23':{'C>T':1},
    'SBS24':{'C>A':1},
    'SBS25':{'T>A':1},
    'SBS26':{'T>C':-1},
    'SBS27':{'T>A':1},
    'SBS29':{'C>A':1},
    'SBS31':{'C>T':1},
    'SBS32':{'C>T':1},
    'SBS33':{'T>C':-1},
    'SBS35':{'C>A':1,'C>T':1},
    'SBS42':{'C>A':1,'C>T':1},
    'SBS45':{'C>A':1},
    'SBS45':{'T>C':1},
    'SBS47':{'T>A':-1,'T>G':1},
    'SBS50':{'C>A':-1},
    'SBS51':{'C>A':1,'C>T':-1,'T>A':1},
    'SBS57':{'T>C':-1,'T>G':-1},
    'SBS60':{'T>G':-1}
     }
    
    categorised_mutations = categorise_mutations(input_mutations, mutation_categories[mutation_type],
                            condensed=True, strand_bias=True)

    _, strand_bias_present, categories_with_TSB = annotate_strand_bias(categorised_mutations)

    if verbose:
        print('TSB found: ', strand_bias_present, categories_with_TSB)
        print('Initial signatures:', input_signatures.columns.tolist())

    signatures_after_TSB_rules = input_signatures
    for signature in input_signatures.columns.tolist():
        if signature in signatures_with_TSB.keys():
            if set(signatures_with_TSB[signature].keys())<=set(categories_with_TSB.keys()):
                # all necessary categories with significant strand bias are present for a given signature
                # now need to check the TSB sign equivalence
                for category in signatures_with_TSB[signature].keys():
                    if signatures_with_TSB[signature][category] != np.sign(categories_with_TSB[category]):
                        if verbose:
                            print('Excluding signature',signature,'as for category',category,'the sign',signatures_with_TSB[signature][category],'differs from',np.sign(categories_with_TSB[category]))
                        signatures_after_TSB_rules.drop(signature, axis=1, inplace=True)
                        # break out of the loop as the signature only needs to be excluded once
                        break
            else:
                signatures_after_TSB_rules.drop(signature, axis=1, inplace=True)
                if verbose:
                    print('Excluding signature',signature,'as',signatures_with_TSB[signature],' is not equal or a subset of ',categories_with_TSB)

    if verbose:
        print('Signatures passing TSB rules:',signatures_after_TSB_rules.columns.tolist())
    
    #print(signatures_after_TSB_rules)
    #Just return the column names to the remaining signatures 
    return signatures_after_TSB_rules.columns



def assign_signatures96(s, genome=0, genome192=0, sig_pcwag=0, par_one=0.01, par_two=0.025):
        
        """
        samples = a single sample
        genome = all samples in 96 context
        genome192 = all samples in 192 context
        sig_pcwaf = the global signatures in 96 context
        
        """
        
        #get the sample's name
        samples = genome.columns[s]
        # get the Transcription bias
        initial_signatures = sig_pcwag.copy()
        #print updates
        print("\n#################################################\n")
        print("Processing sample "+samples+"\n" ) 
        
        if len(genome192) == 0:
            selected_signatures = initial_signatures
        else:
            sample192 = genome192.loc[:,samples]    
            
            # Apply the strand bias rule
            selected_signatures = apply_TSB_rules(sample192, initial_signatures, "SBS", verbose=False)
            
           
        #get the selected signatures indices
        selected_indeces  = get_indeces(list(sig_pcwag.columns), list(selected_signatures))
        sample = np.array(genome.loc[:,samples])[:,np.newaxis]
        #print(len(selected_indeces))
        
        # sigs means the whole signatures
        sigs = sig_pcwag.reset_index()
        sigs, idx, col, _ = sub.read_csv(sigs)
        sigs = sigs.set_index(idx)
        sigs.columns = col
        sigs = np.array(sigs) 
       
        
        #get the initial exposure
        preselected = ["SBS1", "SBS5"]
        preselected_idx = get_indeces(list(col), preselected)
        #exposures, similarity = fit_signatures(sigs[:,0:5], sample)
       
        # Add the signatures in the first round. a = exposures, b= similarity
        a, b = sigopt.add_signatures(sigs, sample, cutoff=par_one, presentSignatures=preselected_idx, toBeAdded=selected_indeces)
        present = np.nonzero(a)[0]
        added_first = get_items_from_index(list(col),list(present))
        #print(samples)
        #print(added_first)
        #print(b)
        #print("\n")
        
        # Add the signatures in the second round. a = exposures, b= similarity
        nonselected_indeces = list(set(list(range(sigs.shape[1])))-set(selected_indeces))
        a,b = sigopt.add_signatures(sigs, sample, cutoff=par_two, presentSignatures=list(present), toBeAdded= nonselected_indeces)
        present = np.nonzero(a)[0]
        added_second = get_items_from_index(list(col),list(present))
        #print(samples)
        #print(added_second)
        #print(b)
        
        
        #get the connected signatures
        connected_singnatures = [["SBS17a", "SBS17b"], ["SBS7a","SBS7b", "SBS7c","SBS7d"], ["SBS2", "SBS13"], ["SBS10a", "SBS10b"]]
        connected = []     
        for i in connected_singnatures:
            #print (i)
            if any(j in added_second for j in i):
                connected = connected + i
        
        
        #get the final signatures
        finalSignatures = list(set(added_second).union(set(connected)))
        finalSinaturesIdx = get_indeces(list(col), finalSignatures)
        finalSinaturesIdx .sort()
        exposures, similarity = sigopt.fit_signatures(sigs[:,finalSinaturesIdx], sample)
        
       
        
        return[finalSinaturesIdx, exposures, list(col), sigs, similarity, genome.columns[s]]
    



def single_sample(csv, csv192="", output="results", par_one=0.01, par_two=0.025, n_cpu=-1):
    
    
    # First, get the path for data files
    paths = cosmic.__path__[0]
    
    if not os.path.exists(output):
        os.makedirs(output)
        
           
    # get the genome
   
    genome, idx, col, mtype = sub.read_csv(csv)
    genome = genome.set_index(idx)
    genome.columns = col
    if csv192 != "":
        # get the 192 context of the genome     
        genome192 = pd.read_csv(csv192, sep=",", index_col = [0,1,2])
        #Check if the 96 context agrees with the 192 context
        if genome.shape[1] != genome192.shape[1]:
            raise Exception('Please the correct format of the 96 and 192 context files')
    else:
        genome192 = ""
    
    
    
    #get the signature files
    sig_pcwag = pd.read_csv(paths+"/input/sigProfiler_SBS_signatures.csv", index_col = [0,1])
    
    
    totalExposures = np.zeros([sig_pcwag.shape[1],genome.shape[1]]) 
    listOfSamples = list(col)
    
    # open a file to profile the signatures
    fh = open(output+"/decomposition profile.csv", "w")
    fh.write("Sample Names, Global NMF Signatures, Similarity\n")
    fh.close()
    
    
    
    # Map the signautures using paraller processing 
    if n_cpu==-1:
        pool = multiprocessing.Pool()
    else:
        pool = multiprocessing.Pool(processes=n_cpu)
    
    pool_assign_signatures96 = partial(assign_signatures96, genome=genome, genome192=genome192, sig_pcwag=sig_pcwag, par_one = par_one, par_two = par_two)
    results = pool.map(pool_assign_signatures96, range(len(genome.columns)))
    
    pool.close()
    pool.join()
    
    # Extract the results generated from pool 
    n=0
    similarities = []
    for result in results:
        
        totalExposures[result[0],n]=result[1]
        listOfSignatures = result[2]
        signatures = pd.DataFrame(result[3])
        
        profile = spss.decomposition_profile(totalExposures[:,n],  result[4], result[2], result[5])
        similarities.append(result[4])
        #write the profiles into file
        fh = open(output+"/decomposition profile.csv", "a")
        fh.write(profile)
        fh.close()
        n+=1
    
    
    
    #prepare the exposures dataframe 
    totalExposures = pd.DataFrame(totalExposures)
    totalExposures = totalExposures.set_index(np.array(listOfSignatures))
    totalExposures.columns = listOfSamples
    totalExposures = totalExposures.rename_axis("Cancer Types", axis="columns")
    #Convert the floats to integers
    totalExposures[listOfSamples] = totalExposures[listOfSamples].applymap(np.int64)
    
    
    
    #remove the rows with all zeros to create the final exposure dataframe
    #exposures = totalExposures.loc[~(totalExposures==0).all(axis=1)]
    
    #presure the signatures dataframe
    signatures = pd.DataFrame(result[3])
    signatures.columns = listOfSignatures
    signatures = signatures.set_index(genome.index)
    signatures = signatures.rename_axis("Signatures", axis="columns")
    
    #Filter the signatures by the exposures rows to get the final signature dataframe
    #signatures = signatures.loc[:,list(exposures.index)] 
    
    #create the probalities 
    probability = sub.probabilities(signatures, totalExposures)
    probability = probability.set_index("Sample")
    probability = probability.rename_axis("", axis="columns")
    
    
    
    # Transform the exposures matrix and insert the cosine similarity column
    exposures = totalExposures.T    
    exposures.insert(loc=0, column='Similarity', value=similarities )
    
    #export results
    signatures.to_csv(output+"/signatures.txt", "\t", index_label=[signatures.columns.name]) 
    exposures.to_csv(output+"/sig_activities.txt", "\t", index_label=["Cancer Types"]) 
    probability.to_csv(output+"/mutation_probabilities.txt", "\t")
    
    
    try:
        #create the dedrogrames
        Y, dn = sub.dendrogram(totalExposures, 0.05, output)
    except:
        pass
    #plot.plotSBS(outputdir+"/signatures.txt", outputdir+"/Signature_plot", "", "96", True)  
    
    # Finish the process
    print ("\n\n Your Job is Completed! Thank you for using Sigprofiler Single Sample.")
    
        
#single_sample("Biliary-AdenoCA.96.csv", csv192="Biliary-AdenoCA.192.csv", output="results", par_one=0.01, par_two=0.025)        
    
    
    
    
    

            




from setuptools import setup



setup(name='sigproSS',
      version='0.0.0.15',
      description=' Decomposes mutational catalogues to mutational signatures',
      url="https://github.com/AlexandrovLab/SigProfilerExtractor.git",
      author='S Mishu Ashiqul Islam',
      author_email='m0islam@ucsd.edu',
      license='UCSD',
      packages=['sigproSS'],
      install_requires=[
          'matplotlib>=2.2.2',
	  'scipy>=1.1.0', 
	  'numpy>=1.14.4', 
	  'pandas>=0.23.4', 
	  'SigProfilerMatrixGenerator>=0.1.20', 
	  'sigProfilerPlotting>=0.1.17', 
	  'sigproextractor>=0.0.5.6', 
	  'statsmodels>=0.9.0',
	  'scikit-learn>=0.20.2',
	  'lxml>=4.3.0',	
      	   ],
      include_package_data=True,      
      zip_safe=False)



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 18 09:38:35 2018

@author: compactmatter
"""

import numpy as np
import scipy
from scipy import optimize
from numpy import linalg as LA
from sigproextractor import subroutines as sub


def cos_sim(a, b):
    
    """Takes 2 vectors a, b and returns the cosine similarity according 
	to the definition of the dot product

    Dependencies: 
    *Requires numpy library. 
    *Does not require any custom function (constructed by me)
    
    Required by:
    * pairwise_cluster_raw
	"""
    dot_product = np.dot(a, b)
    
    norm_a = np.linalg.norm(a)
    norm_b = np.linalg.norm(b)
    #print (dot_product)
    #print ((norm_a * norm_b))
    result = (dot_product / (norm_a * norm_b))
    
    # check for nan value of result
    if result!=result:
        result=0.0
        
    return result


def parameterized_objective2_custom(x, signatures, samples):
    rec = np.dot(signatures, x)
    try:
        y = LA.norm(samples-rec[:,np.newaxis])
    except:
        y = LA.norm(samples-rec)
    
    return y


def constraints1(x, samples):
    sumOfSamples=np.sum(samples, axis=0)
    #print(sumOfSamples)
    #print(x)
    result = sumOfSamples-(np.sum(x))
    #print (result)
    return result


def create_bounds(idxOfZeros, samples, numOfSignatures):
    total = np.sum(samples)
    b = (0.0, float(total))
    lst =[b]*numOfSignatures
    
    for i in idxOfZeros:
        lst[i] = (0.0, 0.0) 
    
    return lst 


def initial_optimization(processes, genomes):
    exposure = np.zeros([processes.shape[1], genomes.shape[1]] )
    # get the total mutations for the given sample
    maxmutation = np.round(np.sum(genomes, axis=0))
    
    for i in range(genomes.shape[1]):
        Gi = genomes[:,i]
        
        #initialize the guess
        x0 = np.random.rand(processes.shape[1], 1)*maxmutation[i]
        x0= x0/np.sum(x0)*maxmutation[i]
        
        #set the bounds and constraints
        bnds = create_bounds([], Gi, processes.shape[1]) 
        cons1 ={'type': 'eq', 'fun': constraints1, 'args':[Gi]} 
        
       
        #the optimization step
        sol = optimize.minimize(parameterized_objective2_custom, x0, args=(processes, Gi),  bounds=bnds, constraints =cons1, tol=1e-30)
        exposure[:,i]=sol.x
        
    return exposure

def remove_signatures(indices, W, exposures, totoalgenomes):
    i = indices
    H = exposures[:,i]
    genomes= totoalgenomes[:,i]
    # make the empty list of the successfull combinations
    successList = [0,[],0] 
    # get the cos_similarity with sample for the oringinal W and H[:,i]
    originalSimilarity= cos_sim(genomes, np.dot(W, H))
    # make the original exposures of specific sample round
    oldExposures = np.round(H)
    
    # set the flag for the while loop
    if len(oldExposures[np.nonzero(oldExposures)])>1:
        Flag = True
    else: 
        Flag = False
        return oldExposures
    # The while loop starts here
    while Flag: 
        
        # get the list of the indices those already have zero values
        if len(successList[1]) == 0:
            initialZerosIdx = list(np.where(oldExposures==0)[0]) 
            #get the indices to be selected
            selectableIdx = list(np.where(oldExposures>0)[0]) 
        elif len(successList[1]) > 1: 
            initialZerosIdx = list(np.where(successList[1]==0)[0]) 
            #get the indices to be selected
            selectableIdx = list(np.where(successList[1]>0)[0])
        else:
            print("iteration is completed")
            #break
        
        
        # get the total mutations for the given sample
        maxmutation = round(np.sum(genomes))
        
        # new signature matrix omiting the column for zero
        #Winit = np.delete(W, initialZerosIdx, 1)
        Winit = W[:,selectableIdx]
        
        # set the initial cos_similarity
        record  = [0.11, []] 
        # get the number of current nonzeros
        l= Winit.shape[1]
        
        for i in range(l):
            #print(i)
            loopSelection = list(range(l))
            del loopSelection[i]
            #print (loopSelection)
            W1 = Winit[:,loopSelection]
           
            
            #initialize the guess
            x0 = np.random.rand(l-1, 1)*maxmutation
            x0= x0/np.sum(x0)*maxmutation
            
            #set the bounds and constraints
            bnds = create_bounds([], genomes, W1.shape[1]) 
            cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genomes]} 
            
            #the optimization step
            sol = optimize.minimize(parameterized_objective2_custom, x0, args=(W1, genomes),  bounds=bnds, constraints =cons1, tol=1e-15)
            
            #print (sol.success)
            #print (sol.x)
            
            #convert the newExposure vector into list type structure
            newExposure = list(sol.x)
            
            #insert the loopZeros in its actual position 
            newExposure.insert(i, 0)
            
            #insert zeros in the required position the newExposure matrix
            initialZerosIdx.sort()
            for zeros in initialZerosIdx:
                newExposure.insert(zeros, 0)
            
            # get the maximum value the new Exposure
            maxcoef = max(newExposure)
            idxmaxcoef = newExposure.index(maxcoef)
            
            newExposure = np.round(newExposure)
            
            
            if np.sum(newExposure)!=maxmutation:
                newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
                
            
            newSample = np.dot(W, newExposure)
            newSimilarity = cos_sim(genomes, newSample) 
             
            difference = originalSimilarity - newSimilarity
            #print(originalSimilarity)
            #print(newSample)
            #print(newExposure)
            #print(newSimilarity)
            
            #print(difference)
            #print (newExposure)
            #print (np.round(H))
            #print ("\n\n")
             
            if difference<record[0]:
                record = [difference, newExposure, newSimilarity]
            
            
        #print ("This loop's selection is {}".format(record))
        
        if record[0]>0.01:   
            Flag=False
        elif len(record[1][np.nonzero(record[1])])==1:
            successList = record 
            Flag=False
        else:
            successList = record
        #print("The loop selection is {}".format(successList))
        
        #print (Flag)
        #print ("\n\n")
    
    #print ("The final selection is {}".format(successList))
    
    if len(successList[1])==0:
        successList = [0.0, oldExposures, originalSimilarity]
    
    #print ("one sample completed")
    return successList[1]



def add_signatures(W, genome, cutoff=0.025, presentSignatures=[], toBeAdded="all"):
     # This function takes an array of signature and a single genome as input, returns a dictionray of cosine similarity, exposures and presence 
     # of signatures according to the indices of the original signature array
    
    if toBeAdded == "all":
        notToBeAdded=[]
    else:               
        #print(len(list(range(W.shape[1]))))
        #print(len(toBeAdded))
        notToBeAdded = list(set(list(range(W.shape[1])))-set(toBeAdded)) # get the indices of the signatures to be excluded
        #print(len(notToBeAdded))
    originalSimilarity = -1 # it can be also written as oldsimilarity
    maxmutation = round(np.sum(genome))
    init_listed_idx = presentSignatures 
    init_nonlisted_idx = list(set(list(range(W.shape[1])))-set(init_listed_idx)-set(notToBeAdded))
    finalRecord = [["similarity place-holder" ], ["newExposure place-holder"], ["signatures place-holder"]] #for recording the cosine difference, similarity, the new exposure and the index of the best signauture
    
    # get the initial original similarity if some signatures are already given to be present
    if len(init_listed_idx)!=0:
        loop_liststed_idx=init_listed_idx
        loop_liststed_idx.sort()
        #print(loop_liststed_idx)
        W1 = W[:,loop_liststed_idx]
       
        #print (W1.shape)
        #initialize the guess
        x0 = np.random.rand(W1.shape[1], 1)*maxmutation
        x0= x0/np.sum(x0)*maxmutation
        
        #set the bounds and constraints
        bnds = create_bounds([], genome, W1.shape[1]) 
        cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]} 
        
        #the optimization step
        sol = optimize.minimize(parameterized_objective2_custom, x0, args=(W1, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
        
        #print(W1)
        #convert the newExposure vector into list type structure
        newExposure = list(sol.x)
        
        # get the maximum value of the new Exposure
        maxcoef = max(newExposure)
        idxmaxcoef = newExposure.index(maxcoef)
        
        newExposure = np.round(newExposure)
        
        # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
        if np.sum(newExposure)!=maxmutation:
            newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
         
        # compute the estimated genome
        est_genome = np.dot(W1, newExposure)
        originalSimilarity = cos_sim(genome[:,0], est_genome)
        finalRecord = [originalSimilarity, newExposure, init_listed_idx]
        
        
    while True:
        bestDifference = -1 
        bestSimilarity = -1
        loopRecord = [["newExposure place-holder"], ["signatures place-holder"], ["best loop signature place-holder"]]
        for sig in init_nonlisted_idx:
            
            
            if len(init_listed_idx)!=0:
                loop_liststed_idx=init_listed_idx+[sig]
                loop_liststed_idx.sort()
                #print(loop_liststed_idx)
                W1 = W[:,loop_liststed_idx]
                #print (W1.shape)
                #initialize the guess
                x0 = np.random.rand(W1.shape[1], 1)*maxmutation
                x0= x0/np.sum(x0)*maxmutation
                
                #set the bounds and constraints
                bnds = create_bounds([], genome, W1.shape[1]) 
                cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]} 
            # for the first time addintion  
            else:
                W1 = W[:,sig][:,np.newaxis]
                #print (W1.shape)        
                #initialize the guess
                x0 = np.ones((1,1))*maxmutation    
            
                #set the bounds and constraints
                bnds = create_bounds([], genome, 1) 
                cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]} 
            
            #the optimization step
            sol = scipy.optimize.minimize(parameterized_objective2_custom, x0, args=(W1, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
            
            #print(W1)
            #convert the newExposure vector into list type structure
            newExposure = list(sol.x)
            
            # get the maximum value of the new Exposure
            maxcoef = max(newExposure)
            idxmaxcoef = newExposure.index(maxcoef)
            
            newExposure = np.round(newExposure)
            
            # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
            if np.sum(newExposure)!=maxmutation:
                newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
             
            # compute the estimated genome
            est_genome = np.dot(W1, newExposure)
            newSimilarity = cos_sim(genome[:,0], est_genome)
            
            difference = newSimilarity - originalSimilarity 
            
            # record the best values so far
            if difference>bestDifference:
                bestDifference = difference
                bestSimilarity = newSimilarity
                loopRecord = [newExposure, W1, sig]  #recording the cosine difference, the new exposure and the index of the best signauture
                #print(newSimilarity)
        
        # 0.01 is the thresh-hold for now 
        if bestSimilarity-originalSimilarity>cutoff:
            originalSimilarity = bestSimilarity
            init_listed_idx.append(loopRecord[2])
            init_nonlisted_idx.remove(loopRecord[2])
            init_listed_idx.sort()
            #print(originalSimilarity)
            finalRecord = [originalSimilarity, loopRecord[0], init_listed_idx, loopRecord[1], genome]
            #print (finalRecord)
            
            if len(init_nonlisted_idx)!= 0:
                
                continue
            else:
                break
        else:
            break
     
    #print("Good so far")
    #print(finalRecord)
    dictExposure= {"similarity":finalRecord[0], "exposures":finalRecord[1], "signatures": finalRecord[2]}  
    addExposure = np.zeros([W.shape[1]])
    addExposure[dictExposure["signatures"]]=dictExposure["exposures"]
    
    return  addExposure, finalRecord[0]



def fit_signatures(W, genome):
    genome = np.array(genome)
    maxmutation = round(np.sum(genome))
     
    
    
    #initialize the guess
    x0 = np.random.rand(W.shape[1], 1)*maxmutation
    x0= x0/np.sum(x0)*maxmutation
    
    #the optimization step
    #set the bounds and constraints
    bnds = create_bounds([], genome, W.shape[1]) 
    cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]}
    
    sol = scipy.optimize.minimize(parameterized_objective2_custom, x0, args=(W, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
    
    #print(W1)
    #convert the newExposure vector into list type structure
    newExposure = list(sol.x)
    
    
    # get the maximum value of the new Exposure
    maxcoef = max(newExposure)
    idxmaxcoef = newExposure.index(maxcoef)
    
    newExposure = np.round(newExposure)
    
    # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
    if np.sum(newExposure)!=maxmutation:
        newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
     
    # compute the estimated genome
    est_genome = np.dot(W, newExposure)
    newSimilarity = cos_sim(genome[:,0], est_genome)
    return (newExposure, newSimilarity)